<?php

// $Id: twinfield_usersync.process.inc,v 1.1 2010/04/28 15:07:44 peterbex Exp $

/**
 * @file
 * Functions for processing user changes, getting them into Twinfield
 *
 * This contains all the Twinfield API interfacing code; creating
 * and destroying connections, performing requests etc.
 */

/**
 * Obtain the code for the current office.
 *
 * This will be required in most messages. If you omit it, there's no
 * guarantee that your changes will end up in the right Twinfield
 * database.
 *
 * @return  Office code (a string)
 */
function twinfield_get_office_code()
{
  return variable_get('twinfield_usersync_office_code');
}

/**
 * Fetch a list of all available offices in Twinfield.
 *
 * @return
 *   An array, with the office codes as key and the office
 *   name/description as value.
 */
function twinfield_office_codes($conn)
{
  $xml = "<list><type>offices</type></list>";
  $res = twinfield_process_xml($conn, $xml);
  $elem = new SimpleXMLElement($res['ProcessXmlStringResult']);
  if ($elem['result'] == '0') {
    $errors = array();
    foreach($elem->xpath('//@msg') as $error)
        $errors[] = $error;;
    $errors = implode('; ', $errors);
    return array('#errors' => $errors);
  }

  $offices = array();
  foreach($elem->xpath('//office') as $office) {
    $offices[(string)$office] = (string)$office['name'];
  }
  return $offices;
}

/**
 * Get the inverted value of twinfield_usersync_field_selection.
 *
 * This simply exchanges keys and values of the array.
 *
 * @return An array of field selections.
 */
function twinfield_usersync_inverted_selection() {
  $selection = variable_get('twinfield_usersync_field_selection', array());

  $result = array();
  foreach($selection as $type => $keys) {
    foreach($keys as $field => $xml_expr) {
      if ($xml_expr !== '-')
        $result[$xml_expr] = $field;
    }
  }
  return $result;
}

/**
 * Write a Drupal user to Twinfield as a customer.
 *
 * @param $conn
 *  The connection to Twinfield, as returned by twinfield_connect()
 * @param $uid
 *  The Drupal uid of the user.
 * @param $user
 *  An array containing the user's details that changed.  Any additional
 *  info which is missing will automatically be loaded.
 * @return
 *  TRUE on success, NULL on failure. On failure, an error is logged
 *  with the watchdog.
 */
function twinfield_write_user_as_customer($conn, $uid, $user=array()) {
  $selection = twinfield_usersync_inverted_selection();

  // Fetch the entire user data
  $user = array_merge((array)user_load(array('uid' => $uid)), $user);
  foreach(content_profile_get_types('names') as $type => $type_name) {
    foreach((array)content_profile_load($type, $uid) as $key => $value) {
      if (!isset($user[$key])) {
        $user[$key] = $value;
      }
    }
  }

  // Stupid CCK fields are in an array. Normalize!
  foreach($user as $key => $value) {
    if (isset($value[0]) && is_array($value[0])) {
      // Fugly hack to get the first value, whatever the key is
      foreach($value[0] as $k => $v) {
        $user[$key] = $v;
        break;
      }
    }
  }

  if (!isset($selection['code']) || !isset($user[$selection['code']])) return;

  $office = twinfield_get_office_code();
  $xml  = "<dimension><type>DEB</type><office>{$office}</office>";
  // Direct fields
  foreach($selection as $expr => $field) {
    $fragments = explode('/', $expr);
    if (sizeof($fragments) === 1) {
      $xml .= "<{$fragments[0]}>";
      $xml .= htmlspecialchars($user[$field]);
      $xml .= "</{$fragments[0]}>";
    }
  }

  // Nesting level 1 (or 2, with aggregate)
  $aggregates = array('address' => 'addresses', 'bank' => 'banks');
  foreach(array('address', 'bank', 'financials') as $tagtype) {
    $tag_open = false;
    foreach($selection as $expr => $field) {
      $fragments = explode('/', $expr);
      if ($fragments[0] ===  $tagtype) {
        if (!$tag_open) {
          if (isset($aggregates[$tagtype]))
            $xml .= "<{$aggregates[$tagtype]}>";
          $xml .= "<{$tagtype}>";
          $tag_open = true;
        }
        $xml .= "<{$fragments[1]}>";
        $xml .= htmlspecialchars($user[$field]);
        $xml .= "</{$fragments[1]}>";
      }
    }
    if ($tag_open) {
      $xml .= "</{$tagtype}>";
      if (isset($aggregates[$tagtype]))
        $xml .= "</{$aggregates[$tagtype]}>";
    }
  }
  $xml .= "</dimension>";

  $res = twinfield_process_xml($conn, $xml);
  $elem = new SimpleXMLElement($res['ProcessXmlStringResult']);
  if ($elem['result'] == '0') {
    $errors = array();
    foreach($elem->xpath('//@msg') as $error)
        $errors[] = $error;;
    $errors = implode('; ', $errors);
    watchdog(
      'twinfield-usersync',
      'Could not save user data! User name: "%name", uid: %uid. Reason: "%reason"',
      array('%name' => $user['name'], '%uid' => $user['uid'], '%reason' => $errors),
      WATCHDOG_ERROR
    );
    return NULL;
  }
  return TRUE;
}

/**
 * Connect and login, then connect to cluster and return connection or NULL in case of errors.
 *
 * @return
 *   NULL in case of failure, an array with cluster and session
 *   in case of success.  This array should be treated as an
 *   opaque data object and passed as-is to all functions which
 *   accept a Twinfield connection.
 *   Don't forget to call twinfield_disconnect() when you no
 *   longer need the connection.
 */
function twinfield_connect() {
  $init = soapclient_init_client(
    'https://login.twinfield.com/webservices/session.asmx?wsdl',
    true,
    array('response_headers' => true)
  );
  if ($init['#error'] !== false) {
    watchdog(
      'twinfield-usersync',
      'Could not connect to Twinfield server! Error message: "%errmsg"',
      array('%errmsg' => $init['#error']),
      WATCHDOG_ERROR
    );
    return NULL;
  }
  
  $session = $init['#return'];
  $credentials = variable_get('twinfield_usersync_credentials', array());
  $result = $session->call('Logon', $credentials);
  if ($result['#error']) {
    watchdog(
      'twinfield-usersync',
      'Error while talking to Twinfield server! Error message: "%errmsg"',
      array('%errmsg' => $result['#error']),
      WATCHDOG_ERROR
    );
    return NULL;
  }

  switch ($result['#return']['LogonResult']) {
    case 'Ok':
      if ($result['#return']['nextAction'] !== 'None') {
        watchdog(
          'twinfield-usersync',
          'Twinfield needs an extra action to complete login. Action required: %action',
          array('%action' => $result['#return']['nextAction']),
          WATCHDOG_ERROR
        );
        // If we got here, we have an error
        return NULL;
      }
      break;
    case 'Blocked':
      watchdog(
        'twinfield-usersync',
        'Twinfield login (temporarily) blocked because of system maintenance',
        array(),
        WATCHDOG_WARNING
      );
      return NULL;
      break;
    default:
      watchdog(
        'twinfield-usersync',
        'Twinfield login failed.  Reason: %reason',
        array('%reason' => $result['#return']['LogonResult']),
        WATCHDOG_ERROR
      );
      return NULL;
      break;
  }

  // If we got here, everything is OK and we can continue connecting to a cluster
  $session_url = $result['#return']['cluster'].'/webservices/processxml.asmx?wsdl';
//  $session_url = 'http://localhost:8080/';
  $h = array('headers' => '<Header xmlns="http://www.twinfield.com/"><SessionID>'.$result['#headers']['SessionID'].'</SessionID></Header>');
  $init_session = soapclient_init_client($session_url, true, $h);
  if ($init_session['#error'] !== false) {
    watchdog(
      'twinfield-usersync',
      'Could not connect to Twinfield application cluster! Error message: "%errmsg"',
      array('%errmsg' => $init_session['#error']),
      WATCHDOG_ERROR
    );
    return NULL;
  }

  return array('cluster' => $init_session['#return'], 'session' => $session);
}

/**
 * Abandon the connection. Needed for clean disconnect.
 *
 * @param $conn
 *   The connection, as returned by twinfield_connect().
 */
function twinfield_disconnect($conn)
{
  $conn['session']->call('Abandon');
}

function twinfield_process_xml($conn, $xml)
{
  $res = $conn['cluster']->call('ProcessXmlString', array('xmlRequest' => $xml));
  if ($res['#error'] !== false )
  {
    watchdog(
      'twinfield-usersync',
      'Error while talking to Twinfield web app! Error message: "%errmsg". XML source: "%source"',
      array('%errmsg' => $res['#error'], '%source' => $xml),
      WATCHDOG_ERROR
    );
    throw new Exception();
  }
  return $res['#return'];
}

?>
